/* Our windowsnapper class */
objectdef windowsnapper
{
	/* Late initialization method, as opposed to using the object constructor. We want to  
	   create the object before we initialize it.
        */
	method Go()
	{
		/* Registering events before we attach ensures that the event exists, to attach to */
		LavishScript:RegisterEvent[On Deactivate]
		LavishScript:RegisterEvent[On Activate]

		/* Attach our methods to the events we registered for */
		Event[On Deactivate]:AttachAtom[This:OnDeactivate]
		Event[On Activate]:AttachAtom[This:OnActivate]

		/* Spit out startup spam */
		echo \agWindowSnapper: Window will snap to ${SnapPos} with size ${SnapSize}
		echo To change snap position, enter command (e.g.) \aysnappos 123,456
		echo To change snap size, enter command (e.g.) \aysnapsize 640x480
		echo WindowSnapper also accepts -size 640x480 and -pos 123,456 as parameters, such as
		echo \ayrun windowsnapper -size 640x480 -pos 123,456
	}

	/* When Store is called, we populate these */
	variable int X
	variable int Y
	variable int Width
	variable int Height
	variable string Frame

	/* This tracks whether we're the active window or not. Without it, window size and position could
           get messy. */
	variable bool Active=TRUE

	variable bool Ignore=FALSE

	/* TRUE when windowsnapper is in Draggable mode, for dragging to set the snap-to size and position */
	variable bool Draggable=FALSE

	/* This is our snap-to size and position. When not the active window, we use these. */
	variable string SnapSize=320x240
	variable string SnapPos=800,0

	/* Grabs the X-position from SnapPos */
	member:int SnapX()
	{
		return ${Arg[1,${SnapPos}]}
	}

	/* Grabs the Y-position from SnapPos */
	member:int SnapY()
	{
		return ${Arg[2,${SnapPos}]}
	}

	/* Grabs the width from SnapSize */
	member:int SnapWidth()
	{
		return ${SnapSize.Token[1,x]}
	}

	/* Grabs the height from SnapSize */
	member:int SnapHeight()
	{
		return ${SnapSize.Token[2,x]}
	}

	/* Store the current display position and size for later use */
	method Store()
	{
		X:Set[${Display.Window.X}]
		Y:Set[${Display.Window.Y}]
		Width:Set[${Display.ViewableWidth}]
		Height:Set[${Display.ViewableHeight}]
		Frame:Set["${Display.Window.Frame}"]
	}

	/* Restore the stored display position and size */
	method Restore()
	{
		squelch windowpos -stealth ${X},${Y}
		windowframe -stealth ${Frame}
		squelch windowsize -rescale -viewable ${Width}x${Height}
	}

	/* Restore the stored display position and size */
	method RestoreFull()
	{
		squelch windowpos ${X},${Y}
		squelch windowsize -viewable ${Width}x${Height}
		windowframe ${Frame}
	}

	/* Toggles draggable mode ... (comments ftw) */
	method ToggleDraggable()
	{
		if ${Draggable}
			This:EndDraggable
		else
			This:BeginDraggable
	}

	/* Enters draggable mode, allowing dragging to set the snap-to size and position */
	method BeginDraggable()
	{
		if ${Draggable}
			return
		Event[On Deactivate]:DetachAtom[This:OnDeactivate]
		Event[On Activate]:DetachAtom[This:OnActivate]
		Draggable:Set[TRUE]
		This:Store
		This:SnapTo
		windowframe -stealth thick				
	}

	/* Completes dragging operation, setting the final snap-to size and position and restoring the full size */
	method EndDraggable()
	{
		if !${Draggable}
			return
		Draggable:Set[FALSE]
		snapsize "${Display.Window.Width}x${Display.Window.Height}"
		snappos "${Display.Window.X},${Display.Window.Y}"
		This:RestoreFull
		Event[On Deactivate]:AttachAtom[This:OnDeactivate]
		Event[On Activate]:AttachAtom[This:OnActivate]
	}

	/* Snaps the window TO the snap-to size and position, with always on top */
	method SnapTo()
	{
		windowframe -stealth none
		squelch windowsize -rescale -viewable ${SnapSize}
		squelch windowpos -stealth ${SnapPos}
		windowvisibility -noactivate alwaysontop
	}

	/* Snaps the window back to full size, and removes always on top */
	method SnapBack()
	{
		This:Restore
		windowvisibility foreground
	}

	/* Handler for the "On Deactivate" event, which fires when the window is no longer the active one */
	method OnDeactivate()
	{
		if ${Ignore}
			return
		if ${Active}
		{
			This:Store
			This:SnapTo
			Active:Set[FALSE]
		}
	}
	
	/* Handler for the "On Activate" event, which fires when the window is once again the active one */
	method OnActivate()
	{
		Ignore:Set[FALSE]
		if !${Active}
		{
			This:SnapBack
			Active:Set[TRUE]
		}
	}
	
}

/* Declare our windowsnapper object, called Snapper */
variable windowsnapper Snapper

/* Create a command called snapsize to change the size. The command is here until the script is terminated. */
atom(global) snapsize(string size)
{
	Snapper.SnapSize:Set["${size}"]
		echo \agWindowSnapper: Window will snap to ${Snapper.SnapPos} with size ${Snapper.SnapSize}
	if !${Snapper.Active}
		squelch windowsize -rescale -viewable ${Snapper.SnapSize}
}

/* Create a command called snappos to change the position. The command is here until the script is terminated. */
atom(global) snappos(string position)
{
	Snapper.SnapPos:Set["${position}"]
		echo \agWindowSnapper: Window will snap to ${Snapper.SnapPos} with size ${Snapper.SnapSize}
	if !${Snapper.Active}
		squelch windowpos -stealth ${Snapper.SnapPos}
}

/* Create a command called snapignore to ignore the next deactivation, until reactivated (again to change your mind) */
atom(global) snapignore()
{
	if ${Snapper.Ignore:Toggle}
		echo \agWindowSnapper: Ignoring next deactivation
	else
		echo \agWindowSnapper: NOT Ignoring next deactivation
}

/* Create a command called snapdrag to turn on, turn off, or toggle Dragging mode */
atom(global) snapdrag(string onofftoggle)
{
	switch ${onofftoggle}
	{
	case on
		Snapper:BeginDraggable
		break
	case toggle
		Snapper:ToggleDraggable
		break
	case off
	default
		Snapper:EndDraggable
		break
	}
}

/* Create a command called snapdock to allow docking of another session's snap-to, adjacent to this session's snap-to */
atom(global) snapdock(string session, string where)
{
	variable string pos

	switch ${where}
	{
	case top
		pos:Set["${Snapper.SnapX},${Math.Calc[${Snapper.SnapY}-${Snapper.SnapHeight}]}"]
		relay "${session}" "snappos ${pos};snapsize ${Snapper.SnapSize}"		
		break
	case bottom
		pos:Set["${Snapper.SnapX},${Math.Calc[${Snapper.SnapY}+${Snapper.SnapHeight}]}"]		
		relay "${session}" "snappos ${pos};snapsize ${Snapper.SnapSize}"		
		break
	case left
		pos:Set["${Math.Calc[${Snapper.SnapX}-${Snapper.SnapWidth}]},${Snapper.SnapY}"]		
		relay "${session}" "snappos ${pos};snapsize ${Snapper.SnapSize}"		
		break
	case right
		pos:Set["${Math.Calc[${Snapper.SnapX}+${Snapper.SnapWidth}]},${Snapper.SnapY}"]		
		relay "${session}" "snappos ${pos};snapsize ${Snapper.SnapSize}"		
		break
	default
		echo Unknown dock position, try top, bottom, left or right
		break
	}
}

/* Entry point for the script. All parameters passed to it will be put in an array called Params */
function main(... Params)
{  
	/* Process parameters */
	variable int p=1
	for (${p}<=${Params.Size} ; p:Inc)
	{
		/* Switch does a string comparison, so parameter processing is easy */
		switch ${Params[${p}]}
		{
		case -size
			/* we got the -size switch, so if there's another parameter, set Snapper.SnapSize to the new value */
			p:Inc
			if ${p}<=${Params.Size}
				Snapper.SnapSize:Set["${Params[${p}]}"]
			break
		case -pos
			/* we got the -pos switch, so if there's another parameter, set Snapper.SnapPos to the new value */
			p:Inc
			if ${p}<=${Params.Size}
				Snapper.SnapPos:Set["${Params[${p}]}"]
			break
		}
	}

	/* Create a bind for snapignore toggling */
	bind snapignore CTRL+ALT+P snapignore
	bind snapdragtoggle CTRL+ALT+O snapdrag toggle
	
	/* Parameters have been processed, let's get going. */	 
	Snapper:Go
	
	/* Now just keep the script running, so our functionality doesn't go away */
	while 1
	{
		waitframe
	}
}

atom atexit()
{
	bind -delete snapignore
	bind -delete snapdragtoggle
}
